/**
 * This file is part of my seminar project.
 * Following code, along with other files in this project,
 * realizes the visualization of A* algorithm.
 * 
 * @author Teon Banek
 */
package hr.fer.seminar;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.SwingUtilities;

/**
 * This class is a {@link Thread} that performs the A* pathfinding algorithm.
 * 
 * @author Teon Banek
 * 
 */
public class Pathfinder extends Thread {

	private Set<Tile> closedSet = new LinkedHashSet<>();
	private Set<Tile> openSet = new LinkedHashSet<>();
	private Map<Tile, Tile> cameFrom = new HashMap<Tile, Tile>();

	private Tile start;
	private Tile goal;
	private TileMap tileMap;

	private int[] gScore;
	private int[] hScore;
	private int[] fScore;

	private volatile int delay;

	/**
	 * Flag: Can the algorithm go diagonally on the grid.
	 */
	private boolean moveDiagonally = false;

	/**
	 * Flag: should the algorithm fast forward steps or wait for user input.
	 */
	private volatile boolean fastForwarding = false;

	/**
	 * Flag: is the thread running.
	 */
	private volatile boolean running = true;

	/**
	 * Flag: should the next step in algorithm be performed.
	 */
	private volatile boolean nextStep = true;

	/**
	 * @param start
	 *            - starting {@link Tile}.
	 * @param goal
	 *            - goal {@link Tile}.
	 * @param tileMap
	 *            - {@link TileMap}.
	 */
	public Pathfinder(Tile start, Tile goal, TileMap tileMap,
			boolean moveDiagonally, int delay) {
		super();
		this.start = start;
		this.goal = goal;
		this.tileMap = tileMap;
		this.moveDiagonally = moveDiagonally;
		this.delay = delay;

		gScore = new int[tileMap.getMapWidth() * tileMap.getMapHeight()];
		hScore = new int[tileMap.getMapWidth() * tileMap.getMapHeight()];
		fScore = new int[tileMap.getMapWidth() * tileMap.getMapHeight()];
		for (int i = 0; i < tileMap.getMapHeight() * tileMap.getMapWidth(); i++) {
			gScore[i] = 0;
			hScore[i] = Integer.MAX_VALUE;
			fScore[i] = Integer.MAX_VALUE;
		}
		
		openSet.add(start);
		int startNode = start.getY() * tileMap.getMapWidth() + start.getX();
		gScore[startNode] = 0;
		hScore[startNode] = calculateHeuristic(start.getX(), start.getY(),
				goal.getX(), goal.getY());
		fScore[startNode] = gScore[startNode] + hScore[startNode];
	}

	@Override
	public void run() {
		while (running) {
			if (fastForwarding || nextStep) {
				if (findPath() != PathfinderStatus.CONTINUE) {
					running = false;
				}
				nextStep = false;
			}
			try {
				Thread.sleep(100);
			} catch (InterruptedException ignorable) {
			}
		}
	}

	/**
	 * Finds and reconstructs the shortest path (if possible).
	 * 
	 * @return {@link PathfinderStatus}
	 */
	public PathfinderStatus findPath() {
		do {
			if (openSet.isEmpty()) {
				return PathfinderStatus.NO_PATH_FOUND;
			}
			if (step() == PathfinderStatus.PATH_FOUND) {
				return PathfinderStatus.PATH_FOUND;
			}
			Tile nextTile = findLowestF();
			if (nextTile == null) {
				return PathfinderStatus.NO_PATH_FOUND;
			}
			int fVal = fScore[nextTile.getY() * tileMap.getMapWidth()
					+ nextTile.getX()];
			SwingUtilities.invokeLater(new modifyTile(nextTile, tileMap, true,
					false, false, fVal));
			if (fastForwarding) {
				try {
					Thread.sleep(delay);
				} catch (InterruptedException ignorable) {
				}
			}
		} while (fastForwarding && running);

		return PathfinderStatus.CONTINUE;
	}

	/**
	 * Perform one step of the path finding algorithm.
	 * 
	 * @return {@link PathfinderStatus}.
	 */
	private PathfinderStatus step() {
		Tile current = findLowestF();
		if (current == goal) {
			reconstructPath(current);
			return PathfinderStatus.PATH_FOUND;
		}

		openSet.remove(current);
		closedSet.add(current);

		SwingUtilities.invokeLater(new modifyTile(current, tileMap, false,
				true, false, fScore[current.getY() * tileMap.getMapWidth()
						+ current.getX()]));

		checkNeighbors(getNeighbors(current), current);

		return PathfinderStatus.CONTINUE;
	}

	/**
	 * Reconstruct the path from given point.
	 * 
	 * @param current
	 *            - last point the pathfinder got to.
	 */
	private void reconstructPath(Tile current) {
		Tile tmp = current;
		while (tmp != start && cameFrom.containsKey(tmp)) {
			SwingUtilities.invokeLater(new modifyTile(tmp, tileMap, false,
					false, true, fScore[tmp.getY() * tileMap.getMapWidth()
							+ tmp.getX()]));
			tmp = cameFrom.get(tmp);
		}
	}

	/**
	 * Analyze the neighboring nodes and choose the best one.
	 * 
	 * @param neighbors
	 *            {@link Set} of neighboring {@link Tile}.
	 * @param current
	 *            {@link Tile} the pathfinder is on.
	 */
	private void checkNeighbors(Set<Tile> neighbors, Tile current) {
		for (Tile n : neighbors) {
			if (closedSet.contains(n) || n.equals(current)) {
				continue;
			}
			int nIndex = n.getY() * tileMap.getMapWidth() + n.getX();
			int curIndex = current.getY() * tileMap.getMapWidth()
					+ current.getX();
			int tentativeG = gScore[curIndex] + costBetween(current, n);
			boolean tentativeBetter = false;
			if (openSet.contains(n) == false) {
				openSet.add(n);
				hScore[nIndex] = calculateHeuristic(n.getX(), n.getY(),
						goal.getX(), goal.getY());
				tentativeBetter = true;
			} else if (tentativeG < gScore[nIndex]) {
				tentativeBetter = true;
			}
			if (tentativeBetter) {
				cameFrom.put(n, current);
				gScore[nIndex] = tentativeG;
				fScore[nIndex] = gScore[nIndex] + hScore[nIndex];
				SwingUtilities.invokeLater(new modifyTile(n, tileMap, false,
						false, false, fScore[nIndex]));
			}
		}
	}

	/**
	 * Calculate cost between given neighboring nodes.
	 * 
	 * @param current
	 *            {@link Tile} where pathfinder is on.
	 * @param n
	 *            neighbor {@link Tile}.
	 * @return double value cost.
	 */
	private int costBetween(Tile current, Tile n) {

		// Uncomment if different diagonal movement cost desired.
		// if (isDiagonallyAdjacent(n.getX(), n.getY(), current.getX(),
		// current.getY())) {
		// return 14;
		// }

		return 10;
	}

	/**
	 * Creates a {@link Set} of neighboring tiles.
	 * 
	 * @param current
	 *            {@link Tile} where pathfinder is on.
	 * @return {@link Set} of {@link Tile}.
	 */
	private Set<Tile> getNeighbors(Tile current) {
		Set<Tile> neighbors = new HashSet<>();
		int curX = current.getX();
		int curY = current.getY();

		for (int x = curX - 1; x <= curX + 1; x++) {
			for (int y = curY - 1; y <= curY + 1; y++) {
				if (x == curX && y == curY) {
					continue;
				}
				boolean diagNeighbor = isDiagonallyAdjacent(x, y, curX, curY);
				boolean nextToWall = isNextToWall(x, y, curX, curY);
				Tile n = tileMap.getTile(x, y);
				if (n != null && n.isBlocking() == false) {
					if (diagNeighbor && (!moveDiagonally || nextToWall)) {
						continue;
					}
					neighbors.add(n);
				}
			}
		}
		return neighbors;
	}

	/**
	 * Test if given coordinates are diagonally adjacent.
	 * 
	 * @param x1
	 *            - first x coordinate.
	 * @param y1
	 *            - first y coordinate.
	 * @param x2
	 *            - second x coordinate.
	 * @param y2
	 *            - second y coordinate.
	 * @return true if diagonally adjacent.
	 */
	private boolean isDiagonallyAdjacent(int x1, int y1, int x2, int y2) {
		boolean diagNeighbor = ((x1 == x2 - 1 || x1 == x2 + 1) && (y1 == y2 - 1 || y1 == y2 + 1));
		return diagNeighbor;
	}

	/**
	 * Tests if the given coordinates are next to a wall from current
	 * perspective, thus preventing diagonal movement.
	 * 
	 * @param x
	 *            - target x coordinate.
	 * @param y
	 *            - target y coordinate.
	 * @param curX
	 *            - current x coordinate.
	 * @param curY
	 *            - current y coordinate.
	 * @return true if wall obstructs diagonal movement.
	 */
	private boolean isNextToWall(int x, int y, int curX, int curY) {
		if (isDiagonallyAdjacent(x, y, curX, curY) == false) {
			return false;
		}
		Tile t = tileMap.getTile(curX, y);
		if (t != null && t.isBlocking()) {
			return true;
		}
		t = tileMap.getTile(x, curY);
		if (t != null && t.isBlocking()) {
			return true;
		}
		return false;
	}

	/**
	 * Calculates for each node on open lists the cost F.
	 * 
	 * @return lowest cost {@link Tile}.
	 */
	private Tile findLowestF() {
		double minF = Double.POSITIVE_INFINITY;
		Tile minTile = null;
		for (Tile t : openSet) {
			if (closedSet.contains(t)) {
				continue;
			}
			int loc = t.getY() * tileMap.getMapWidth() + t.getX();
			if (fScore[loc] < minF || t.equals(goal)) {
				minF = fScore[loc];
				minTile = t;
			}
			if (fScore[loc] == minF) {
				minTile = breakTie(t, minTile);
			}
		}
		return minTile;
	}

	/**
	 * Breaks ties of F score.
	 * 
	 * @param t1
	 *            - first tile.
	 * @param t2
	 *            - second tile.
	 * @return
	 */
	private Tile breakTie(Tile t1, Tile t2) {
		// Tie breaking based on lower heuristic score.
		int h1 = calculateHeuristic(t1.getX(), t1.getY(), goal.getX(),
				goal.getY());
		int h2 = calculateHeuristic(t2.getX(), t2.getY(), goal.getX(),
				goal.getY());
		if (h1 < h2) {
			return t1;
		} else {
			return t2;
		}
	}

	/**
	 * Calculates the heuristic for given nodes. Manhattan distance is used.
	 * 
	 * @return
	 */
	private int calculateHeuristic(int x1, int y1, int x2, int y2) {
		// Manhattan distance multiplied by lowest cost of moving to a tile.
		int h = (Math.abs(x2 - x1) + Math.abs(y2 - y1)) * 10;
		return h;
	}

	/**
	 * @return the fastForwarding
	 */
	public boolean isFastForwarding() {
		return fastForwarding;
	}

	/**
	 * @param fastForwarding
	 *            the fastForwarding to set
	 */
	public void setFastForwarding(boolean fastForwarding) {
		this.fastForwarding = fastForwarding;
	}

	/**
	 * Performs the jobs of modifying a tile. Instance of this class is given to
	 * EventQueue thread.
	 * 
	 * @author Teon Banek
	 * 
	 */
	private static class modifyTile implements Runnable {
		private Tile tile;
		private TileMap tileMap;
		private boolean next;
		private boolean closed;
		private boolean path;
		private int fVal;

		/**
		 * The constructor takes a {@link Tile} to be modified and modifies its
		 * flags by given values.
		 * 
		 * @param tile
		 *            to be modified.
		 * @param tileMap
		 *            - grid based graph where the given tile is located.
		 * @param next
		 *            - flag: is tile the next to be put in closed set.
		 * @param closed
		 *            - flag: is tile on closed set.
		 * @param path
		 *            - flag: is tile part of the shortest path.
		 * @param fVal
		 *            - calculated F score of the tile.
		 */
		public modifyTile(Tile tile, TileMap tileMap, boolean next,
				boolean closed, boolean path, int fVal) {
			super();
			this.tile = tile;
			this.tileMap = tileMap;
			this.next = next;
			this.closed = closed;
			this.path = path;
			this.fVal = fVal;
		}

		@Override
		public void run() {
			tile.setLabelF(fVal);
			tile.setPath(path);
			tile.setClosed(closed);
			tile.setNext(next);
			tileMap.updateUI();
		}
	}

	/**
	 * @param running
	 *            the running to set
	 */
	public void setRunning(boolean running) {
		this.running = running;
	}

	/**
	 * @param nextStep
	 *            the nextStep to set
	 */
	public void setNextStep(boolean nextStep) {
		this.nextStep = nextStep;
	}

	/**
	 * @param delay
	 *            the delay to set
	 */
	public void setDelay(int delay) {
		this.delay = delay;
	}
}
